from collections import OrderedDict
import json
import os
import logging
from typing import Dict, List, Optional

import anndata as ad
import cellxgene_census
import pandas as pd

from cas.cxg_utils import download_dataset_with_id
from cas.file_utils import read_anndata_file


logging.basicConfig(level=logging.INFO)

LABELSET_COLUMN = "CELL LABELSET NAME"
CELL_LABEL_COLUMN = "CELL TYPE TERM"
CL_TERM_COLUMN = "CL TERM"
EVIDENCE_COLUMN = "EVIDENCE"
MARKER_GENES_COLUMN = "MARKER GENES"
SYNONYM_COLUMN = "SYNONYMS"
CATEGORIES_COLUMN = "CATEGORIES"


def read_spreadsheet(file_path: str, sheet_name: Optional[str]):
    """
    Read the specific sheet from the Excel file into a pandas DataFrame.

    Args:
        file_path (str): Path to the Excel file.
        sheet_name (str, optional): Target sheet name. If not provided, reads the first sheet.

    Returns:
        tuple: Tuple containing metadata (dict), column names (list), and raw data (pd.DataFrame).
    """
    # Read the specific sheet or the first sheet if sheet_name is not provided
    if sheet_name:
        spreadsheet_df = pd.read_excel(file_path, sheet_name=sheet_name, header=None)
    else:
        spreadsheet_df = pd.read_excel(file_path, header=None)
    # Extract meta data (first 8 rows)
    meta_data = dict(spreadsheet_df.iloc[:8, 0:2].to_records(index=False))
    # Extract column names (9th row)
    column_names = spreadsheet_df.iloc[8, :].tolist()
    # Extract raw data (from 10th row onwards)
    raw_data = spreadsheet_df.iloc[9:, :]
    raw_data.columns = column_names

    return meta_data, column_names, raw_data


def get_cell_ids(dataset: ad.AnnData, labelset: str, cell_label: str) -> List[str]:
    """
    Get cell IDs from an AnnData dataset based on a specified labelset and cell label.

    Args:
        dataset (ad.AnnData): AnnData dataset.
        labelset (str): Labelset to filter.
        cell_label (str): Cell label to match.

    Returns:
        List[str]: List of cell IDs.
    """
    cell_label_lower = str(cell_label).lower()
    return dataset.obs.index[
        dataset.obs[labelset].astype(str).str.lower() == cell_label_lower
        ].tolist()


def calculate_labelset_rank(input_list: List[str]) -> Dict[str, int]:
    """
    Assign ranks to items in a list.

    Args:
        input_list (List[str]): The list of items.

    Returns:
        Dict[str, int]: A dictionary where keys are items from the input list and
        values are their corresponding ranks (0-based).

    """
    return {item: rank for rank, item in enumerate(input_list)}


def spreadsheet2cas(
    spreadsheet_file_path: str,
    sheet_name: Optional[str],
    anndata_file_path: str,
    labelset_list: Optional[List[str]],
    output_file_path: str,
):
    """
    Convert a spreadsheet to Cell Annotation Schema (CAS) JSON.

    Args:
        spreadsheet_file_path (str): Path to the spreadsheet file.
        sheet_name (Optional[str]): Target sheet name in the spreadsheet. Can be a string or None.
        anndata_file_path: The path to the AnnData file.
        labelset_list (Optional[List[str]]): List to determine the rank of labelsets in spreadsheet.
        output_file_path (str): Output CAS file name.
    """
    meta_data_result, column_names_result, raw_data_result = read_spreadsheet(
        spreadsheet_file_path, sheet_name
    )

    matrix_file_id = (
        meta_data_result["CxG LINK"].rstrip("/").split("/")[-1].split(".")[0]
    )
    if not anndata_file_path:
        download_dataset_with_id(matrix_file_id)
    dataset_anndata = read_anndata_file(anndata_file_path)

    labelsets = OrderedDict()

    # metadata
    cas = {
        "matrix_file_id": matrix_file_id,
        "cellannotation_schema_version": "TBA",
        "cellannotation_timestamp": "TBA",
        "cellannotation_version": "TBA",
        "cellannotation_url": meta_data_result["CxG LINK"],
        "author_name": "TBA",
        "author_contact": "TBA",
        "orcid": "TBA",
        "annotations": [],
        "labelsets": [],
    }

    # annotations
    stripped_data_result = raw_data_result.map(
        lambda x: x.strip() if isinstance(x, str) else x
    )
    for index, row in stripped_data_result.iterrows():
        labelset = row[LABELSET_COLUMN]
        cell_label = row[CELL_LABEL_COLUMN]
        cell_ontology_term_id = row[CL_TERM_COLUMN]
        rationale = row[EVIDENCE_COLUMN]
        marker_gene_evidence = row[MARKER_GENES_COLUMN]
        synonyms = row[SYNONYM_COLUMN]
        category_fullname = row[CATEGORIES_COLUMN]

        # labelsets.add(labelset)
        labelsets[labelset] = None

        anno = {
            "labelset": labelset,
            "cell_label": cell_label,
            "cell_fullname": cell_label,
            "cell_ontology_term_id": cell_ontology_term_id,
            "cell_ontology_term": cell_label,
            "cell_ids": get_cell_ids(dataset_anndata, labelset, cell_label),
            "rationale": rationale,
            "rationale_dois": meta_data_result["PAPER DOI"],
            "marker_gene_evidence": marker_gene_evidence,
            "synonyms": synonyms,
            "category_fullname": category_fullname,
            "category_cell_ontology_exists": "TBA",
            "category_cell_ontology_term_id": "TBA",
            "category_cell_ontology_term": "TBA",
        }
        cas.get("annotations").append(anno)

    # labelsets
    labelset_rank_dict = calculate_labelset_rank(labelset_list if labelset_list else list(labelsets.keys()))
    for labelset, rank in labelset_rank_dict.items():
        cas.get("labelsets").append({"name": labelset, "description": "TBA", "rank": str(rank)})

    # Write the JSON data to the file
    with open(output_file_path, "w") as json_file:
        json.dump(cas, json_file, indent=2)



"""
cxg_utils.py

This module provides utility functions for working with AnnData datasets in the context of the CellxGene Census library.
"""
import os
import logging

import cellxgene_census


def download_dataset_with_id(dataset_id: str):
    """
    Download an AnnData dataset with a specified ID.

    Args:
        dataset_id (str): ID of the dataset.

    """
    anndata_file_path = f"{dataset_id}.h5ad"
    # Check if the file already exists
    # Currently using dataset id as file names for the downloads
    if os.path.exists(anndata_file_path):
        print(f"File '{anndata_file_path}' already exists. Skipping download.")
    else:
        logging.info(f"Downloading dataset with ID '{dataset_id}'...")
        cellxgene_census.download_source_h5ad(dataset_id, to_path=anndata_file_path)
        logging.info(f"Download complete. File saved at '{anndata_file_path}'.")



import argparse
import pathlib
import sys
import warnings

from cas.anndata_conversion import merge
from cas.flatten_data_to_anndata import flatten
from cas.populate_cell_ids import populate_cell_ids
from cas.spreadsheet_to_cas import spreadsheet2cas
from cas.anndata_to_cas import anndata2cas
from cas.validate import validate as schema_validate

warnings.filterwarnings("always")


def main():
    parser = argparse.ArgumentParser(
        prog="cas", description="Cell Type Annotation Tools cli interface."
    )
    subparsers = parser.add_subparsers(help="Available ctat actions", dest="action")

    create_merge_operation_parser(subparsers)
    create_flatten_operation_parser(subparsers)
    create_spreadsheet2cas_operation_parser(subparsers)
    create_anndata2cas_operation_parser(subparsers)
    create_populate_cells_operation_parser(subparsers)
    create_schema_validation_operation_parser(subparsers)

    args = parser.parse_args()

    if args.action == "merge":
        json_file_path = args.json
        anndata_file_path = args.anndata
        output_file_path = args.output
        validate = args.validate

        if anndata_file_path == output_file_path:
            raise ValueError("--anndata and --output cannot be the same")

        merge(json_file_path, anndata_file_path, validate, output_file_path)
    elif args.action == "flatten":
        args = parser.parse_args()
        json_file_path = args.json
        anndata_file_path = args.anndata
        output_file_path = args.output
        validate = args.validate

        if anndata_file_path == output_file_path:
            raise ValueError("--anndata and --output cannot be the same")

        flatten(json_file_path, anndata_file_path, validate, output_file_path)
    elif args.action == "spreadsheet2cas":
        args = parser.parse_args()
        spreadsheet_file_path = args.spreadsheet
        sheet_name = args.sheet
        anndata_file_path = args.anndata
        labelsets = args.labelsets
        output_file_path = args.output

        spreadsheet2cas(
            spreadsheet_file_path, sheet_name, anndata_file_path, labelsets, output_file_path
        )
    elif args.action == "anndata2cas":
        args = parser.parse_args()
        anndata_file_path = args.anndata
        labelsets = args.labelsets
        output_file_path = args.output
        include_hierarchy = args.hierarchy

        anndata2cas(
            anndata_file_path, labelsets, output_file_path, include_hierarchy
        )
    elif args.action == "populate_cells":
        args = parser.parse_args()
        json_file_path = args.json
        anndata_file_path = args.anndata
        labelsets = None
        if "labelsets" in args and args.labelsets:
            labelsets = [item.strip() for item in str(args.labelsets).split(",")]
        populate_cell_ids(json_file_path, anndata_file_path, labelsets)
    elif args.action == "validate":
        args = parser.parse_args()
        schema = args.schema
        data_file_path = args.data
        schema_validate(schema, data_file_path)


def create_merge_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --json      : Path to the CAS JSON schema file.
    --anndata   : Path to the AnnData file. Ideally, the location will be specified by a resolvable path in the CAS file.
    --validate  : Perform validation checks before writing to the output AnnData file.
    --output    : Output AnnData file name (default: output.h5ad).

    Usage Example:
    --------------
    cd src
    python -m cas merge --json path/to/CAS_schema.json --anndata path/to/input_anndata.h5ad --validate --output path/to/output.h5ad
    """
    parser_merge = subparsers.add_parser(
        "merge",
        description="The CAS and AnnData merge parser",
        help="Test if CAS can be merged to the AnnData and merges if possible.",
    )
    parser_merge.add_argument(
        "--json", required=True, help="Path to the CAS JSON schema file."
    )
    parser_merge.add_argument(
        "--anndata", required=True, help="Path to the AnnData file."
    )
    # TODO find a better argument name and Help message.
    parser_merge.add_argument(
        "-v",
        "--validate",
        action="store_true",
        help="Perform validation checks before writing to the output AnnData file.",
    )
    parser_merge.add_argument(
        "--output",
        default="output.h5ad",
        help="Output AnnData file name (default: output.h5ad).",
    )
    parser_merge.set_defaults(validate=False)


def create_flatten_operation_parser(subparsers):
    """
        Command-line Arguments:
    -----------------------
    --json      : Path to the CAS JSON schema file.
    --anndata   : Path to the AnnData file. Ideally, the location will be specified by a resolvable path in the CAS file.
    --validate  : Perform validation checks before flattening to AnnData file.
    --output    : Output AnnData file name (default: output.h5ad).

    Usage Example:
    --------------
    cd src
    python -m cas flatten --json path/to/json_file.json --anndata path/to/anndata_file.h5ad --output path/to/output_file.h5ad
    """
    parser_flatten = subparsers.add_parser(
        "flatten",
        description="Flattens all content of CAS annotations to an AnnData file.",
        help="Flattens all content of CAS annotations to obs key:value pairs. "
        "Flattens all other content to key_value pairs in uns.",
    )

    parser_flatten.add_argument("--json", required=True, help="Input JSON file path")
    parser_flatten.add_argument(
        "--anndata", required=True, help="Input AnnData file path"
    )
    parser_flatten.add_argument(
        "-v",
        "--validate",
        action="store_true",
        help="Perform validation checks before writing to the output AnnData file.",
    )
    parser_flatten.add_argument(
        "--output",
        help="Output AnnData file name (default: output.h5ad)",
        default="output.h5ad",
    )
    parser_flatten.set_defaults(validate=False)


def create_spreadsheet2cas_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --spreadsheet   : Path to the spreadsheet file.
    --sheet         : Target sheet name in the spreadsheet.
    --anndata       : Path to the AnnData file. If not provided, AnnData will be downloaded using CxG LINK in
                    spreadsheet.
    --labelsets     : List to determine the rank of labelsets in spreadsheet. If not provided, ranks will be determined using the order of CELL LABELSET NAME.
    --output        : Output CAS file name (default: output.json).

    Usage Example:
    --------------
    cd src
    python -m cas spreadsheet2cas --spreadsheet path/to/spreadsheet.xlsx --sheet sheet_name --labelsets item1 item2 item3 --output path/to/output_file.json
    """

    parser_spreadsheet2cas = subparsers.add_parser(
        "spreadsheet2cas",
        description="Converts a spreadsheet to CAS JSON.",
        help="Converts a spreadsheet to Cell Annotation Schema (CAS) JSON.",
    )

    parser_spreadsheet2cas.add_argument(
        "--spreadsheet", required=True, help="Path to the spreadsheet file."
    )
    parser_spreadsheet2cas.add_argument(
        "--sheet", required=False, help="Target sheet name in the spreadsheet."
    )
    parser_spreadsheet2cas.add_argument(
        "--anndata",
        default=None,
        help="Path to the AnnData file. If not provided, AnnData will be downloaded using CxG LINK in spreadsheet.",
    )
    parser_spreadsheet2cas.add_argument(
        "--labelsets",
        default=None,
        nargs='+',
        help="List to determine the rank of labelsets in spreadsheet. If not provided ranks will be determined using "
             "order of CELL LABELSET NAME.",
    )
    parser_spreadsheet2cas.add_argument(
        "--output",
        default="output.json",
        help="Output CAS file name (default: output.json).",
    )


def create_anndata2cas_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --anndata       : Path to the AnnData file.
    --labelsets     : List of labelsets.
    --output        : Output CAS file name (default: output.json).
    --hierarchy     : Flag indicating whether to include hierarchy in the output.


    Usage Example:
    --------------
    cd src
    python -m cas anndata2cas --anndata path/to/anndata.h5ad --labelsets item1 item2 item3 --output
    path/to/output_file.json
    """
    parser_anndata2cas = subparsers.add_parser(
        "anndata2cas",
        description="Converts an anndata to CAS JSON.",
        help="Converts an anndata to Cell Annotation Schema (CAS) JSON.",
    )

    parser_anndata2cas.add_argument(
        "--anndata",
        required=True,
        help="Path to the AnnData file.",
    )
    parser_anndata2cas.add_argument(
        "--labelsets",
        required=True,
        nargs='+',
        help="List of labelsets.",
    )
    parser_anndata2cas.add_argument(
        "--output",
        default="output.json",
        help="Output CAS file name (default: output.json).",
    )
    parser_anndata2cas.add_argument(
        "--hierarchy",
        action="store_true",
        help="Include hierarchy in the output.",
    )


def create_populate_cells_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --json      : Path to the CAS JSON schema file.
    --anndata   : Path to the AnnData file. Ideally, the location will be specified by a resolvable path in the CAS file.
    --labelsets : List of labelsets to update with IDs from AnnData. If value is not provided, rank '0' labelset is used.

    Usage Example:
    --------------
    cd src
    python -m cas populate_cells --json path/to/json_file.json --anndata path/to/anndata_file.h5ad --labelsets Cluster,Supercluster
    """
    parser_populate = subparsers.add_parser(
        "populate_cells",
        description="The CAS cell IDs population operation.",
        help="Checks for alignment between obs key:value pairs in AnnData file and labelset:cell_label pairs in CAS for some specified list of labelsets. If they are aligned, updates cell_ids in CAS.",
    )

    parser_populate.add_argument("--json", required=True, help="Input JSON file path")
    parser_populate.add_argument(
        "--anndata", required=True, help="Input AnnData file path"
    )
    parser_populate.add_argument(
        "--labelsets",
        help="List of labelsets to update with IDs from AnnData",
        default="",
    )


def create_schema_validation_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --schema    : One of 'base', 'bican' or 'cap'. Identifies the CAS schema to validate data against.
    --data   : Path to the data file (or folder) to validate

    Usage Example:
    --------------
    cd src
    python -m cas validate --schema bican --data path/to/file
    """
    parser_validate = subparsers.add_parser(
        "validate",
        description="The CAS file structure validator",
        help="Test if given CAS is compatible with the CAS schema.",
    )

    parser_validate.add_argument(
        "--schema", required=True, help="Schema name: one of 'base', 'bican' or 'cap'"
    )
    parser_validate.add_argument(
        "--data",
        required=True,
        help="Path to the data file (or folder) to validate",
        type=pathlib.Path,
    )


if __name__ == "__main__":
    main()



import json
import os
import warnings
from importlib import resources

from cas_schema import schema_validator, schemas

warnings.filterwarnings("always")

SCHEMA_FILE_MAPPING = {
    "base": "general_schema.json",
    "cap": "CAP_schema.json",
    "bican": "BICAN_schema.json",
}


def validate(schema_name: str, data_path: str):
    """
    Validates all instances in data_path against the given schema.
    Assumes all *.json files in the test_dir should validate against the schema.
    Logs all validation errors and throws an exception if any of the test files is invalid.
    Parameters:
        schema_name: One of 'base', 'bican' or 'cap'. Identifies the CAS schema to validate data against.
        data_path: Path to the data file (or folder) to validate
    """
    schema_name = str(schema_name).strip().lower()
    if schema_name not in SCHEMA_FILE_MAPPING:
        raise Exception("Schema name should be one of 'base', 'bican' or 'cap'")
    if not os.path.exists(data_path):
        raise Exception("Please provide a valid 'data_path': {}".format(data_path))

    schema_file = resources.files(schemas) / SCHEMA_FILE_MAPPING[schema_name]
    with schema_file.open("rt") as f:
        schema = json.loads(f.read())

    result = schema_validator.validate(
        schema, SCHEMA_FILE_MAPPING[schema_name], data_path
    )
    if not result:
        raise Exception("Validation Failed")

    return True



import json
from typing import List

from cas.file_utils import read_anndata_file
from cas.spreadsheet_to_cas import calculate_labelset_rank, get_cell_ids
from cas.accession.hash_accession_manager import HashAccessionManager


def anndata2cas(
    anndata_file_path: str,
    labelsets: List[str],
    output_file_path: str,
    include_hierarchy: bool,
):
    """
    Convert an AnnData file to Cell Annotation Schema (CAS) JSON.

    Args:
        anndata_file_path (str): Path to the AnnData file.
        labelsets (List[str]): List of labelsets.
        output_file_path (str): Output CAS file name.
        include_hierarchy (bool): Flag indicating whether to include hierarchy in the output.
    """

    anndata = read_anndata_file(anndata_file_path)

    labelset_dict = calculate_labelset(anndata.obs, labelsets)

    cas = generate_cas_metadata(dict(anndata.uns))

    parent_cell_look_up = generate_cas_annotations(
        anndata, cas, include_hierarchy, labelset_dict
    )

    generate_cas_labelsets(cas, labelsets)

    add_parent_cell_hierarchy(cas, include_hierarchy, parent_cell_look_up)

    # Write the JSON data to the file
    with open(output_file_path, "w") as json_file:
        json.dump(cas, json_file, indent=2)


def add_parent_cell_hierarchy(cas, include_hierarchy, parent_cell_look_up):
    """
    Adds parent cell hierarchy information to the CAS dictionary.

    Args:
        cas (dict): The CAS dictionary.
        include_hierarchy (bool): Flag indicating whether to include hierarchy in the output.
        parent_cell_look_up (dict): Dictionary containing parent cell information.

    Returns:
        None
    """
    if include_hierarchy:
        for key, value in parent_cell_look_up.items():
            for inner_key, inner_value in parent_cell_look_up.items():
                if value != inner_value and value.get("cell_ids").issubset(
                    inner_value.get("cell_ids")
                ):
                    value.update(
                        {
                            "parent": inner_key,
                            "p_accession": inner_value.get("accession"),
                        }
                    )

        annotation_list = cas.get("annotations")
        for annotation in annotation_list:
            annotation.update(
                {
                    "parent_cell_set_name": parent_cell_look_up.get(
                        annotation.get("cell_label")
                    ).get("parent")
                }
            )
            annotation.update(
                {
                    "parent_cell_set_accession": parent_cell_look_up.get(
                        annotation.get("cell_label")
                    ).get("p_accession")
                }
            )


def generate_cas_annotations(anndata, cas, include_hierarchy, labelset_dict):
    """
    Generates CAS annotations and returns parent_cell_look_up dictionary that is calculated during annotation
    generation.

    Args:
        anndata (AnnData): The AnnData object.
        cas (dict): The CAS dictionary.
        include_hierarchy (bool): Flag indicating whether to include hierarchy in the output.
        labelset_dict (dict): Dictionary containing labelsets.

    Returns:
        parent_cell_look_up (dict): A dictionary containing information about parent cell lookups.
    """
    accession_manager = HashAccessionManager()
    parent_cell_look_up = {}
    for k, v in labelset_dict.items():
        for label in v:
            labelset = k
            cell_label = label
            cell_ontology_term_id = anndata.obs[anndata.obs[k] == cell_label][
                "cell_type_ontology_term_id"
            ].iloc[0]
            cell_ontology_term = anndata.obs[anndata.obs[k] == cell_label][
                "cell_type"
            ].iloc[0]
            cell_ids = get_cell_ids(anndata, labelset, label)
            cell_set_accession = accession_manager.generate_accession_id(
                cell_ids=cell_ids
            )
            # TODO None values will be calculated later on
            rationale = None
            rationale_dois = None
            marker_gene_evidence = None
            synonyms = None
            category_fullname = None
            category_cell_ontology_exists = None
            category_cell_ontology_term_id = None
            category_cell_ontology_term = None

            if include_hierarchy:
                if cell_label in parent_cell_look_up:
                    parent_cell_look_up[cell_label].get("cell_ids").update(cell_ids)
                else:
                    parent_cell_look_up[cell_label] = {
                        "cell_ids": set(cell_ids),
                        "accession": cell_set_accession,
                    }

            anno_init = {
                "labelset": labelset,
                "cell_label": cell_label,
                "cell_fullname": cell_label,
                "cell_set_accession": cell_set_accession,
                "cell_ontology_term_id": cell_ontology_term_id,
                "cell_ontology_term": cell_ontology_term,
                "cell_ids": cell_ids,
                "rationale": rationale,
                "rationale_dois": rationale_dois,
                "marker_gene_evidence": marker_gene_evidence,
                "synonyms": synonyms,
                "category_fullname": category_fullname,
                "category_cell_ontology_exists": category_cell_ontology_exists,
                "category_cell_ontology_term_id": category_cell_ontology_term_id,
                "category_cell_ontology_term": category_cell_ontology_term,
            }
            # Exclude keys with None values
            cas.get("annotations").append(
                {k: v for k, v in anno_init.items() if v is not None}
            )
    return parent_cell_look_up


def generate_cas_labelsets(cas, labelsets):
    """
    Generates CAS labelsets and updates the provided CAS dictionary.

    Args:
        cas (dict): The CAS dictionary.
        labelsets (list): List of labelsets.

    Returns:
        None
    """
    # labelsets
    labelset_rank_dict = calculate_labelset_rank(labelsets)
    for labelset, rank in labelset_rank_dict.items():
        cas.get("labelsets").append(
            {"name": labelset, "description": "", "rank": str(rank)}
        )


def generate_cas_metadata(uns):
    """
    Generates CAS metadata based on the provided 'uns' dictionary.

    Args:
        uns (dict): The 'uns' dictionary containing metadata.

    Returns:
        dict: The generated CAS metadata dictionary.
    """
    # TODO None values will be calculated later on
    matrix_file_id = None
    cellannotation_schema_version = uns["schema_version"]
    cellannotation_timestamp = None
    cellannotation_version = None
    cellannotation_url = None
    author_name = None
    author_contact = None
    orcid = None
    cas_init = {
        "matrix_file_id": matrix_file_id,
        "cellannotation_schema_version": cellannotation_schema_version,
        "cellannotation_timestamp": cellannotation_timestamp,
        "cellannotation_version": cellannotation_version,
        "cellannotation_url": cellannotation_url,
        "author_name": author_name,
        "author_contact": author_contact,
        "orcid": orcid,
        "annotations": [],
        "labelsets": [],
    }
    # Exclude keys with None values
    cas = {k: v for k, v in cas_init.items() if v is not None}
    return cas


def calculate_labelset(obs, labelsets):
    """
    Calculates labelset dictionary based on the provided observations.

    Args:
        obs (DataFrame): DataFrame containing observations.
        labelsets (List[str]): List of labelsets.

    Returns:
        dict: The calculated labelset dictionary.
    """
    labelset_dict = {}
    for item in labelsets:
        labelset_dict.update({item: set(obs[item])})
    return labelset_dict



import pathlib
from setuptools import setup, find_packages


# The directory containing this file
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="cas-tools",
    version="0.0.1.dev24",
    description="Cell Annotation Schema tools.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/cellannotation/cas-tools",
    author="",
    license="Apache-2.0 license",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={'': 'src'},
    packages=find_packages(where='src', exclude=("test*",)),
    include_package_data=True,
    install_requires=["anndata==0.10.3", "dataclasses_json", "pandas",
                      "ruamel.yaml", "jsonschema"],
    entry_points={
        "console_scripts": [
            "cas=cas.__main__:main",
        ]
    },
)



#!/usr/bin/env python3

"""
Script Description:
This script processes and integrates information from a JSON file and an AnnData (Annotated Data) file,
creating a new AnnData object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

Key Features:
1. Parses command-line arguments for input JSON file, input AnnData file, and output file.
2. Reads and processes the input JSON file and AnnData file.
3. Updates the AnnData object with information from the JSON annotations and root keys.
4. Writes the modified AnnData object to a specified output file.
"""

from cas.anndata_conversion import test_compatibility
from cas.file_utils import read_anndata_file, read_json_file

LABELSET_NAME = "name"

LABELSET = "labelset"

LABELSETS = "labelsets"

ANNOTATIONS = "annotations"

CELL_IDS = "cell_ids"

CELL_LABEL = "cell_label"


def is_list_of_strings(var):
    """
    Check if a value is a list of strings.

    Parameters:
        var (list or any): The value to be checked.

    Returns:
        bool: True if the value is a list containing only string elements,
              False otherwise.

    """
    return isinstance(var, list) and all(isinstance(item, str) for item in var)


def flatten(json_file_path, anndata_file_path, validate, output_file_path):
    """
     Processes and integrates information from a JSON file and an AnnData (Annotated Data) file, creating a new AnnData
     object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

    Args:
        json_file_path: The path to the CAS json file.
        anndata_file_path: The path to the AnnData file.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
        output_file_path: Output AnnData file name.
    """
    input_json = read_json_file(json_file_path)
    input_anndata = read_anndata_file(anndata_file_path)

    flatten_cas_object(input_json, input_anndata, validate, output_file_path)


def flatten_cas_object(input_json, input_anndata, validate, output_file_path):
    """
     Processes and integrates information from a JSON file and an AnnData (Annotated Data) file, creating a new AnnData
     object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

    Args:
        input_json: CAS json file.
        input_anndata: The AnnData object.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
        output_file_path: Output AnnData file name.
    """
    # obs
    annotations = input_json[ANNOTATIONS]

    parent_cell_ids = collect_parent_cell_ids(input_json)

    for ann in annotations:
        cell_ids = []
        if CELL_IDS in ann and ann[CELL_IDS]:
            cell_ids = ann[CELL_IDS]
        elif (
            "cell_set_accession" in ann and ann["cell_set_accession"] in parent_cell_ids
        ):
            cell_ids = list(parent_cell_ids[ann["cell_set_accession"]])

        for k, v in ann.items():
            if k == CELL_IDS or k == LABELSET:
                continue
            # if k == CELL_LABEL:
            #     key = ann[LABELSET]
            else:
                key = f"{ann[LABELSET]}--{k}"

            value = v
            if isinstance(v, list):
                non_dict_v = [value for value in v if not isinstance(value, dict)]
                value = ", ".join(sorted(non_dict_v))
                if len(v) > len(non_dict_v):
                    print("WARN: dict values are excluded on field '{}'".format(key))
                input_anndata.obs[key] = ""

            for index_to_insert in cell_ids:
                input_anndata.obs.at[index_to_insert, key] = value
    # uns
    uns_json = {}
    root_keys = list(input_json.keys())
    root_keys.remove(ANNOTATIONS)
    for key in root_keys:
        value = input_json[key]
        if is_list_of_strings(value):
            uns_json[key] = ", ".join(sorted(value))
        elif isinstance(value, str):
            uns_json[key] = value
        else:
            for labelset in value:
                for k, v in labelset.items():
                    if k == LABELSET_NAME:
                        continue
                    new_key = f"{labelset.get(LABELSET_NAME, '')}--{k}"
                    uns_json.update({new_key: v})
    input_anndata.uns.update(uns_json)
    # Close the AnnData file to prevent blocking
    input_anndata.file.close()
    input_anndata.write(output_file_path)


def collect_parent_cell_ids(cas):
    parent_cell_ids = dict()

    labelsets = sorted(cas[LABELSETS], key=lambda x: int(x["rank"]))
    for labelset in labelsets:
        ls_annotations = [
            ann for ann in cas[ANNOTATIONS] if ann["labelset"] == labelset["name"]
        ]

        for ann in ls_annotations:
            if "parent_cell_set_accession" in ann:
                cell_ids = set()
                if CELL_IDS in ann and ann[CELL_IDS]:
                    cell_ids = set(ann[CELL_IDS])
                elif (
                    "cell_set_accession" in ann
                    and ann["cell_set_accession"] in parent_cell_ids
                ):
                    cell_ids = parent_cell_ids[ann["cell_set_accession"]]

                if ann["parent_cell_set_accession"] in parent_cell_ids:
                    parent_cell_ids[ann["parent_cell_set_accession"]].update(cell_ids)
                else:
                    parent_cell_ids[ann["parent_cell_set_accession"]] = set(cell_ids)

    return parent_cell_ids
